;  DISK SEQUENCES:
;
;  2 separate types of disk sequences are provided:
;
;      Diablo/Floppy sequences --
;           a. Commands: read, write, compare single sector.
;           b. Sector address 15 bits (16th bit stripped off).
;           c. Platters: F and R.
;
;      Phoenix sequences --
;           a. Commands: read, write, compare single sectors;
;              copy, verify, multi-sector write.
;           b. Sector address 24 bits (although existing code
;              always sets high 8-bits to 0).
;           c. Platters: 16, specfied by low 4 bits of device
;              address:
;                          x0 -- platter 10 (removable)
;                          x1 -- platter 00 (1st fixed)
;                          x2 -- platter 01 (2nd fixed)
;                             .
;                             .
;                             .
;                          xF -- platter 0E (15th fixed)
;
;      The disks indicate their type to the OS by the value
;      returned in response to a 'disk initialization' command:
;
;                Cx -- Diablo/Floppy
;                Dx -- Phoenix
;
;      BASIC-2 indicates that the low 4-bits of the device
;      address specify the disk platter by using device type
;      'D'.  The OS translates ACTDEV (device type & address)
;      so that device type is 'D' iff disk is a Phoenix type.
;      Translation is as follows:
;
;           disk = Diablo or Floppy
;                T/D10 ==> /B10
;                T/D11 ==> /310
;
;           disk = Phoenix
;                T/310, F/310, F/B10 ==> /D11
;                T/B10, R/310, R/B10 ==> /D10
; CONSIDERATIONS:
;      1. Control write loop with RB for Phoenix.

; COMMAND EQUATES:

$READ      EQU       00H      ; READ SECTOR
$EXTEND    EQU       20H      ; EXTENDED COMMAND (NEXT BYTE HAS COMMAND CODE)
$EXTENDH   EQU       2H       ; (HIGH 4BITS)
$WRITE     EQU       40H      ; WRITE SECTOR
$COMPARE   EQU       80H      ; COMPARE SECTOR

$COPY      EQU       01H      ; COPY SECTORS
$VERIFY    EQU       12H      ; VERIFY SECTORS
$MSWRTB    EQU       10H      ; BEGIN OF MULTI-SECTOR WRITE
$MSWRTE    EQU       11H      ; END OF MULTI-SECTOR WRITE
$FORMAT    EQU       02H      ; FORMAT PLATTER
$TRACK     EQU       18H      ; FORMAT TRACK
$STATUS    EQU       16H      ; DEVICE STATUS
$TURNOFF   EQU       17H      ; TURN OFF RETRY

;      DWRITE:  WRITE ONE SECTOR TO DISK
;               (WRITEB ON 2200T, 4652)
;
;      DWRITEB: WRITE 256 BYTES OF DATA & LRC TO DISK
;               (DWRITE ON 2200T, 4190)
;
;           TRANSFER RATE:
;
;              DIABLO/FLOPPY TRANSFER RATE = 15.7 USEC/CHAR
;                                    (4.3MS FOR 256 BYTES)
;
;              PHOENIX TRANSFER RATE = 5.4 USEC/CHAR
;                               (1.4MS FOR 256 BYTES)
;
;                     (ACTUALLY SLOWER DUE TO DISK PROCESSOR)
;
;              2200 TRANSFER RATE = 18.2 USEC/ CHAR.
;                               (4.6 MS FOR 256 BYTES)


DWRITE:
        MOV     BYTE PTR DS:[RETRY_COUNT],0    ;INIT RETRY COUNT
DWRITE2:
        INC     BYTE PTR DS:[RETRY_COUNT]      ;INCREMENT RETRY COUNT
DWRITE1:
        CALL    CMNDINIT                       ;INITIAL DISK CONTROLLER
        MOV     BYTE PTR DS:[COMMAND1],$WRITE
        CALL    DISKCMND        ;SEND WRITE COMMAND WITH SECTOR ADDR TO DISK
        CMP     AL,0
        JZ      DWRITE4         ;BRANCH IF NO ECHO ERROR

        CMP     BYTE PTR DS:[RETRY_COUNT],16
        JB      DWRITE2         ;BRANCH IF < 16 RETRIES
        JMP     ERROR_94

DWRITE4:
        CALL    AKN
;
        TEST    AL,08H          ; AL =  echo character ( 0Ch indicates access is denied)
        JZ      DWRITE9

        MOV     AX,STACK_BUFFER ;RESET STACK
        MOV     SP,AX

        CALL    UNHOG_DISK
        CALL    DELAY5

        JMP     PROCESS_NEXT1   ;HOLD AND PROCESS NEXT

DWRITE9:
        TEST    AL,04H
        JNZ     ERROR_94F

        CALL    DWRITEB

;       mov     dx,280h
;       in      al,dx

        MOV     CX,15
        CALL    WAITIBS
        JC      ERROR_92F
        TEST    AL,4
        JZ      DWRITE3
        CMP     BYTE PTR DS:[RETRY_COUNT],3
        JNE     DWRITE2

DWRITE3:
        CALL    AKN1E
;       JMP     UNHOG_DISK
        RET

ERROR_94F:
        JMP     ERROR_94
ERROR_92F:
        JMP     ERROR_92

;  DATA TRANSFER
;

DWRITEB:
        MOV     BL,0                       ;INIT LRC COUNT
        MOV     DX,WORD PTR DS:[ACTDEV]
        AND     DH,0FH
        CMP     DH,0DH
        JZ      DWRITEP         ;BRANCH IF PHOENIX TYPE DRIVE

; DIABLO/FLOPPY WRITE SEQUENCE (DELAY CONTROLLED)
;

        MOV     CX,256
        MOV     SI,WORD PTR DS:[BUFFER_PTR]
DWRITEB1:
        LODS    BYTE PTR ES:[SI]
        ADD     BL,AL                        ;ADD CHARACTER TO LRC

        CALL    DELAY10                      ;DELAY SINCE DISK ACCEPTS CHARACTERS AT MAX. RATE
        OUT     0080H,AL                     ;STROBE DATA TO DISK

        LOOP    DWRITEB1                     ;BRANCH IF < 256 CHARACTERS SENT

        MOV     AL,BL
        CALL    DELAY10                      ;DELAY SINCE DISK ACCEPTS CHAR. AT RATE OF 1/8USEC
        OUT     0080H,AL                     ;STROBE LRC TO DISK
        JMP     DELAY10

; PHOENIX WRITE SEQUENCE (READY/BUSY CONTROLLED)
;

DWRITEP:
        MOV     CX,256
        MOV     SI,WORD PTR DS:[BUFFER_PTR]
DWRITEP1:
        MOV     DX,2200H
DWRITEP9:
        IN      AL,0084H      ;5
        TEST    AL,20H        ;3
        JZ      DWRITEP2      ;10
        DEC     DX            ;2
        JNZ     DWRITEP9      ;12

        JMP     ERROR_92                     ;BRANCH IF NOT READY WITHIN 3MS.
DWRITEP2:
        LODS    BYTE PTR ES:[SI]
        ADD     BL,AL                        ;ADD CHARACTER TO LRC
        OUT     0080H,AL
        CALL    DELAY5
        CALL    DELAY5

        LOOP    DWRITEP1

        MOV     DX,2200H
DWRITEP8:
        IN      AL,0084H      ;5
        TEST    AL,20H        ;3
        JZ      DWRITEP3      ;10
        DEC     DX            ;2
        JNZ     DWRITEP8      ;12

        JMP     ERROR_92                     ;BRANCH IF NOT READY WITHIN 3MS.

DWRITEP3:
        MOV     AL,BL
        OUT     0080H,AL                     ;STROBE LRC COUNT TO DISK
        CALL    DELAY5
        CALL    DELAY5
        RET

;      DCOMPARE: PERFORM READ-AFTER-WRITE
;

DCOMPARE:
        MOV     BYTE PTR DS:[RETRY_COUNT],0  ;INIT RETRY COUNT
DCMPARE2:
        INC     BYTE PTR DS:[RETRY_COUNT]    ;INCREMENT RETRY COUNT
        CALL    CMNDINIT

        MOV     BYTE PTR DS:[COMMAND1],$COMPARE
        CALL    DISKCMND

        CMP     AL,0
        JE      DCMPARE3                     ;BR, IF NO ECHO ERROR
        CMP     BYTE PTR DS:[RETRY_COUNT],16
        JB      DCMPARE2

        JMP     ERROR_94

DCMPARE3:
        CALL    AKN                          ;GET ACKNOWLEDGEMENT FOR VALID SECTOR ADDRESS
        TEST    AL,08H                       ;AL =  echo character ( 0Ch indicates access is denied)
        JZ      DCMPARE9

        MOV     AX,STACK_BUFFER              ;RESET STACK WHEN DENIED
        MOV     SP,AX

        CALL    UNHOG_DISK
        CALL    DELAY5

        JMP     PROCESS_NEXT1

DCMPARE9:
        TEST    AL,04H
        JNZ     ERROR_94T
        CALL    OBSDISK
        MOV     CX,15
        CALL    WAITIBS                      ;WAIT ACKNOWLEDGEMENT FROM DISK
        JC      ERROR_92T
        TEST    AL,07H
        JZ      DCMPARE5                     ;BR, IF NOT CRC,FORMAT OR SEEK ERROR
DCMPARE4:
        CMP     BYTE PTR DS:[RETRY_COUNT],16
        JB      DCMPARE2

ERROR_99:
        MOV     AL,99H
        JMP     DISKERR

ERROR_92T:
        JMP     ERROR_92
ERROR_94T:
        JMP     ERROR_94

DCMPARE5:
        CALL    DWRITEB                      ;WRITE DATA AND LRC TO DISK
        MOV     CX,15
        CALL    WAITIBS                      ;WAIT FOR RESPONSE FROM DISK (COMPARE RESULT)
        JC      ERROR_92T
        CMP     AL,0
        JNZ     DCMPARE4                     ;BR, IF COMPARE ERROR
        JMP     DWRITE3

; MSWRTBEG: BEGIN OF MULTI-SECTOR WRITE COMMAND ==> DISK
; MSWRTEND: END OF MULTI-SECTOR WRITE COMMAND ==> DISK
;
;      OPERATIONS THAT WRITE SEQUENTIAL SECTORS (E.G. 'SAVE')
;      CAN BE PRECEDED BY A 'BEGIN OF MULTI-SECTOR WRITE'
;      COMMAND AND TERMINATED BY AN 'END OF MULTI-SECTOR
;      WRITE' COMMAND.  THE DISK PROCESSOR (PHOENIX) WILL
;      MULTI-BUFFER THE SECTORS TO BE WRITTEN, WRITING ONLY
;      WHEN: (1) ALL BUFFERS ARE FULL OR
;            (2) 'END M.S. WRITE' COMMAND IS RECEIVED OR
;            (3) DISK IS RESET OR
;            (4) A COMMAND THAT IS NOT A SEQUENTIAL WRITE
;                IS RECEIVED.
;
;      ERRORS OCCURING DURING WRITING MUST EITHER RESET THE
;      DISK OR SEND AN 'END' COMMAND SO THAT THE BUFFERED
;      SECTORS WILL BE WRITTEN.
;
;      DISK ERRORS ARE ACKNOWLEDGED BY 'END MULTI-SECTOR
;      WRITE' SEQUENCE.
;
;      MS WRITE COMMANDS NOT ISSUED IF READ/AFTER/WRITE

MSWRTBEG:
        MOV     BYTE PTR DS:[COMMAND2],$MSWRTB
MSW1:
        MOV     BYTE PTR DS:[RETRY_COUNT],0  ;INIT RETRY COUNT
MSWRT1:
        INC     BYTE PTR DS:[RETRY_COUNT]    ;INCREMENT RETRY COUNT
        CALL    CMNDINIT                     ;DISK INITIALIZATION
        CMP     AL,0D0H
        JNZ     MSWRT4                       ;BR, IF NOT PHOENIX OR DDDD FLOPPY

MSWRT2:
        MOV     AL,0
        OUT     0004H,AL
        OUT     0008H,AL                     ;IOBS = OPERATION IN PROGRESS

        MOV     BYTE PTR DS:[COMMAND1],30H
        MOV     DX,WORD PTR DS:[ACTDEV]
        TEST    DL,0FH
        JZ      MSWRT3                       ;BR, IF DEVICE Dx0

        AND     DL,0FH                       ;CALCULATE PLATTER #
        DEC     DL                           ;D11 ==>00, D12 ==>01, ...
        OR      DL,$EXTEND
        MOV     BYTE PTR DS:[COMMAND1],DL

MSWRT3:
        MOV     AL,BYTE PTR DS:[COMMAND1]    ;COMMNAD/PLATTER ==> DISK
        CALL    OBSECHO
        CMP     AL,0
        JE      MSWRT3A                      ;BR, IF NO ECHO ERROR
        CMP     BYTE PTR DS:[RETRY_COUNT],16
        JB      MSWRT1
        JMP     ERROR_94

MSWRT3A:
        MOV     AL,BYTE PTR DS:[COMMAND2]
        CALL    OBSECHO                      ;M.S. WRITE COMMAND ==> DISK
        CMP     AL,0
        JE      MSWRT3B                      ;BR, IF NO ECHO ERROR

        CMP     AL,60H                       ;ECHO = 60h indicates extended command reject
        JNZ     MSWRT9

        MOV     AX,STACK_BUFFER
        MOV     SP,AX

        CALL    UNHOG_DISK
        CALL    DELAY5

        JMP     PROCESS_NEXT1

MSWRT9:
        CMP     BYTE PTR DS:[RETRY_COUNT],16
        JB      MSWRT1
        JMP     ERROR_94

MSWRT3B:
        MOV     AL,0
        JMP     OBSDISK                      ;START OPERATIOM

MSWRT4:
        MOV     BYTE PTR DS:[WRITE_FLAG],0
        RET


;   END MULTI-SECTOR WRITE ROUTINE
;

MSWRTEND:
        MOV     BYTE PTR DS:[COMMAND2],$MSWRTE ;COMMAND = END OF MULTI-SECTOR WRITE
        CALL    MSW1                         ;DISK INITIALIZATION (START OF OPERATION)
        MOV     DX,WORD PTR DS:[ACTDEV]
        AND     DH,0FH
        CMP     DH,0DH
        JZ      MSWRTEND9                    ;BR, IF PHOENIX OR DDDD FLOPPY
        RET

MSWRTEND9:
        MOV     CX,15
        CALL    WAITIBS                      ;GET ACKNOWEDGEMENT
        JC      ERROR_92R
        TEST    AL,08H
        JNZ     ERROR_99S                    ;BR, IF READ-AFTER-ERITE ERROR
        JMP     AKN1E                        ;CHECK FOR OTHER ERRORS

ERROR_92R:
        JMP     ERROR_92
ERROR_99S:
        JMP     ERROR_99


;      DREAD: READ ONE SECTOR FROM DISK
;             (DLOAD1 ON 2200T, 41A4)
;

DREAD:
        MOV     BYTE PTR DS:[RETRY_COUNT],0  ;INIT RETRY COUNT
DREADE:
        INC     BYTE PTR DS:[RETRY_COUNT]    ;INCREMENT RETRY COUNT
        CALL    CMNDINIT
        MOV     BYTE PTR DS:[COMMAND1],$READ
        CALL    DISKCMND                     ;SEND READ COMMAND WITH SECTOR ADDRESS TO DISK

        CMP     AL,0
        JE      DREADE1                      ;BR, IF NO ECHO ERROR

        CMP     BYTE PTR DS:[RETRY_COUNT],16
        JB      DREADE                       ;BR, IF < 16 RETRIES
        JMP     ERROR_94

DREADE1:
        CALL    AKN                          ;GET ACKNOWLEDGEMENT FOR VALID SECTOR ADDRESS
        TEST    AL,08H                       ;AL =  echo character ( 0Ch indicates access is denied)
        JZ      DREAD9

        MOV     AX,STACK_BUFFER              ;RESET STACK
        MOV     SP,AX

        CALL    UNHOG_DISK
        CALL    DELAY5

        JMP     PROCESS_NEXT1

DREAD9:
        TEST    AL,04H
        JNZ     ERROR_94S
        CALL    OBSDISK                      ;SIGNAL DISK TO CHECK IOB'S TO INSURE NOT RETRYING
        CALL    AKN1                         ;WAIT ACKNOWLEDGEMENT FROM DISK

;       mov     dx,280h
;       in      al,dx

;       DATA TRANSFER
;

DREADB:
        MOV     DI,WORD PTR DS:[BUFFER_PTR]  ;POINT TO BUFFER
        MOV     BL,0
        MOV     CX,256

        MOV     AL,01H
        OUT     0002H,AL            ;SET CPU READY FOR ALLOW INPUT

DREADB1:
        MOV     DX,900H*15

DREAD_AGAIN:
        IN      AL,0084H            ;5
        TEST    AL,02H              ;3
        JZ      DREAD_NOT_OK        ;3 (7+7)
        NOP
        NOP
        NOP
        NOP
        NOP
        NOP
        NOP
        NOP
        IN      AL,0082H            ;5       GET DATA
        STOSB
        ADD     BL,AL               ;3
;       PUSH    AX
;       POP     AX
        LOOP    DREADB1

        MOV     DX,900H*15

DREAD_AGAIN1:
        IN      AL,0084H            ;5
        TEST    AL,02H              ;3
        JZ      DREAD_NOT_OK1       ;3 (7+7)
        NOP
        NOP
        NOP
        NOP
        NOP
        NOP
        NOP
        NOP
        IN      AL,0082H            ;5       GET DATA

        CMP     BL,AL
        JNE     LRCERR                       ;BR, IF LRC WRONG

        MOV     AL,00H
        OUT     0002H,AL            ;SET CPU BUSY

;       JMP     UNHOG_DISK
        RET

DREAD_NOT_OK1:
        DEC     DX                  ;2
        JNZ     DREAD_AGAIN1        ;7+4
        MOV     AL,00H
        OUT     0002H,AL            ;SET CPU BUSY
        JMP     ERROR_92

DREAD_NOT_OK:
        DEC     DX                  ;2
        JNZ     DREAD_AGAIN         ;7+4
        MOV     AL,00H
        OUT     0002H,AL            ;SET CPU BUSY

ERROR_92S:
        JMP     ERROR_92
ERROR_94S:
        JMP     ERROR_94

LRCERR:
        MOV     AL,00H
        OUT     0002H,AL            ;SET CPU BUSY

        CMP     BYTE PTR DS:[RETRY_COUNT],16
        JAE     ERROR_97
        JMP     DREADE

ERROR_97:
        MOV     AL,97H                       ;LRC ERROR
        JMP     DISKERR


; DFORMAT:  FORMAT DISK STATEMENT
;

DFORMAT:
        MOV     BYTE PTR DS:[RETRY_COUNT],0  ;INITIALIZE RETRY COUNT
FORMATD3:
        INC     BYTE PTR DS:[RETRY_COUNT]

        CALL    CMNDINIT
        MOV     BYTE PTR DS:[COMMAND1],$EXTEND
        MOV     BYTE PTR DS:[COMMAND2],$FORMAT
        CALL    DISKCMDC
        CMP     AL,0
        JZ      FORMATD1

        CMP     AL,60H                 ;ECHO = 60h indicates extended command reject
        JNZ     FORMAT9

        MOV     AX,STACK_BUFFER
        MOV     SP,AX

        CALL    UNHOG_DISK
        CALL    DELAY5

        JMP     PROCESS_NEXT1

FORMAT9:
        TEST    AL,0F0H
        JZ      FORMAT3A
        TEST    AL,0FH
        JZ      FORMAT3A
        JMP     ERROR_93

FORMAT3A:
        CMP     BYTE PTR DS:[RETRY_COUNT],16
        JB      FORMATD3
        JMP     ERROR_94

FORMATD1:
        MOV     AL,0
        CALL    OBSDISK
        CALL    AKN1

;       JMP     UNHOG_DISK
        RET

; DVERIFY : Verify Sectors
;

DVERIFY:
        MOV     BYTE PTR DS:[RETRY_COUNT],0    ;INIT RETRY COUNT
VERIFY25:
        INC     BYTE PTR DS:[RETRY_COUNT]
        CALL    CMNDINIT                       ;INITIALIZE DISK
        MOV     DX,WORD PTR DS:[ACTDEV]
        AND     DH,0FH
        CMP     DH,0DH
        JNE     VERIFY21X                      ;OLD TYPE DISK VERIFY
        MOV     BYTE PTR DS:[COMMAND1],$EXTEND ;EXTENDED COMMAND
        MOV     BYTE PTR DS:[COMMAND2],$VERIFY ;COMMAND = VERIFY
        CALL    DISKCMND                       ;/COMMAND/PLATTER/SECTOR ADDRESS ==> DISK
        CMP     AL,0
        JZ      VERIFY6                        ;BR IF COMMAND OK

        CMP     AL,60H                         ;ECHO = 60h indicates extended command reject
        JNZ     VERIFY9

        MOV     AX,STACK_BUFFER
        MOV     SP,AX

        CALL    UNHOG_DISK
        CALL    DELAY5

        JMP     PROCESS_NEXT1

VERIFY21X:
        JMP     VERIFY21

VERIFY9:
        CMP     BYTE PTR DS:[RETRY_COUNT],16
        JB      RETRYVRF                       ;IF COMMAND ERROR & RETRY COUNT < 16 THEN BRANCH
        JMP     ERROR_94

RETRYVRF:
        TEST    AL,0F0H
        JZ      VERIFY25
        TEST    AL,0FH                         ;BR IF ECHO ERROR
        JZ      VERIFY25
        JMP     VERIFY21                       ;BR IF DPU VERIFY NOT SUPPORT
VERIFY6:
        CALL    AKN                            ;ACKNOWLEDGEMENT OF VALID(START) SECTOR ADDRESS

        MOV     AL,BYTE PTR DS:[BP].@ADDR21_5  ;END SECTOR ADDRESS
        MOV     BYTE PTR DS:[ADDR1],AL
        MOV     AL,BYTE PTR DS:[BP].@ADDR22_5
        MOV     BYTE PTR DS:[ADDR2],AL
        MOV     AL,BYTE PTR DS:[BP].@ADDR23_5
        MOV     BYTE PTR DS:[ADDR3],AL

        CALL    DISKCMDA                       ;END ADDRESS ==> DISK
        CALL    AKN                            ;ACKNOWLEDGEMENT OF VALID(END)SECTOR ADDRESS
        MOV     AL,0
        CALL    OBSDISK                        ;START VERIFY
        MOV     CX,15
        CALL    WAITIBS                        ;WAIT IBS --- BREAKPOINT IF NOT READY WITHIN 1/4 MS.
        JC      ERR_92A
        MOV     BYTE PTR DS:[BP].@ADDR31_5,AL  ;1ST ADDRESS
        MOV     CX,15
        CALL    WAITIBS                        ;FOR SET RECEIVER,THE HIGH BYTE SECTOR ADDRESS IS IGNORED IN ANKBPT
        JC      ERR_92A
        MOV     BYTE PTR DS:[BP].@ADDR32_5,AL  ;2ND ADDRESS
        MOV     CX,15
        CALL    WAITIBS
        JC      ERR_92A
        MOV     BYTE PTR DS:[BP].@ADDR33_5,AL  ;3RD ADDRESS
        CALL    AKN1                           ;READ ACKNOWLEDGE
;       JMP     UNHOG_DISK                     ;NO ERROR
        RET

ERR_92A:
        JMP     ERR_92

VERIFY21:
        MOV     AL,BYTE PTR DS:[ADDR1]
        MOV     BYTE PTR DS:[BP].@ADDR31_5,AL
        MOV     AL,BYTE PTR DS:[ADDR2]
        MOV     BYTE PTR DS:[BP].@ADDR32_5,AL
        MOV     AL,BYTE PTR DS:[ADDR3]
        MOV     BYTE PTR DS:[BP].@ADDR33_5,AL
        CALL    GET_BUFFER_PTR                         ;GET BUFFER POINTER
        CALL    DREAD                                  ;DO READ SECTOR ACTION
        CALL    ADD_SECTOR_ADDR                        ;INCREMENT SECTOR ADDRESS
        MOV     AL,BYTE PTR DS:[ADDR3]
        CMP     AL,BYTE PTR DS:[BP].@ADDR23_5
        JNE     VERIFY21                               ;BR, TO READ AGAIN

        MOV     AL,BYTE PTR DS:[ADDR2]
        CMP     AL,BYTE PTR DS:[BP].@ADDR22_5
        JNE     VERIFY21                               ;BR, TO READ AGAIN

        MOV     AL,BYTE PTR DS:[ADDR1]
        CMP     AL,BYTE PTR DS:[BP].@ADDR21_5
        JNE     VERIFY21                                ;BR, TO READ AGAIN

;       JMP     UNHOG_DISK
        RET

;DTRACK: FORMAT DISK TRACK
;

DTRACK:
        MOV     BYTE PTR DS:[RETRY_COUNT],0 ;INIT RETRY COUNT
DTRACK3:
        INC     BYTE PTR DS:[RETRY_COUNT]   ;INCREMENT RETRY COUNT
        CALL    CMNDINIT                    ;INITIALIZATION COMMAND
        MOV     BYTE PTR DS:[COMMAND1],$EXTEND  ;EXTENDED COMMAND
        MOV     BYTE PTR DS:[COMMAND2],$TRACK   ;COMMAND = FORMAT TRACK
        CALL    DISKCMND                    ;SEND FORMAT TRACK COMMAND & SECTOR ADDRESS TO DISK
        CMP     AL,0
        JE      DTRACK1                     ;BRANCH IF NO ECHO ERROR

        CMP     AL,60H                      ;ECHO = 60h indicates extended command reject
        JNZ     DTRACK9

        MOV     AX,STACK_BUFFER
        MOV     SP,AX

        CALL    UNHOG_DISK
        CALL    DELAY5

        JMP     PROCESS_NEXT1

DTRACK9:
        CMP     BYTE PTR DS:[RETRY_COUNT],16
        JB      DTRACK3                     ;BRANCH IF < 16 RETRIES
        JMP     ERROR_94
DTRACK1:
        CALL    AKN                         ;GET ACKNOWLEDGE FOR VALID SECTOR ADDRESS
        MOV     AL,00H
        CALL    OBSDISK                     ;START FORMATTING
        CALL    AKN1                        ;ACKNOWLEDGEMENT OF FORMAT COMPLETE

;       JMP     UNHOG_DISK
        RET

;DSTATUS:  READ DPU STATUS
;

DSTATUS:
        MOV     BYTE PTR DS:[RETRY_COUNT],0 ;INIT RETRY COUNT
DSTATUS3:
        INC     BYTE PTR DS:[RETRY_COUNT]   ;INCREMENT RETRY COUNT
        CALL    CMNDINIT                    ;INITIALIZATION COMMAND
        MOV     BYTE PTR DS:[COMMAND1],$EXTEND  ;EXTENDED COMMAND
        MOV     BYTE PTR DS:[COMMAND2],$STATUS  ;COMMAND = READ STATUS
        CALL    DISKCMDC                    ;SEND COMMAND TO DISK
        CMP     AL,0
        JE      DSTATUS1                    ;BRANCH IF COMMAND OK

        CMP     AL,60H
        JNE     DSTATUS9

        MOV     AX,STACK_BUFFER
        MOV     SP,AX
        CALL    UNHOG_DISK
        CALL    DELAY5
        JMP     PROCESS_NEXT1

DSTATUS9:
        CMP     BYTE PTR DS:[RETRY_COUNT],16
        JB      DSTATUS3                    ;BRANCH IF < 16 RETRIES
        JMP     ERROR_94

DSTATUS1:
        MOV     AL,00H
        CALL    OBSDISK                     ;START READING STATUS

        MOV     CX,15
        CALL    WAITIBS                     ;GET MESSAGE COUNT
        JC      ERROR_92Y

        XOR     DX,DX
        MOV     DL,AL

        CALL    GET_BUFFER_PTR
        MOV     BX,WORD PTR DS:[BUFFER_PTR]

        MOV     BYTE PTR ES:[BX],AL          ;COUNT
        INC     BX
DSTATUS2:
        MOV     CX,15
        CALL    WAITIBS                      ;GET STATUS BYTE
        JC      ERROR_92Y
        MOV     BYTE PTR ES:[BX],AL          ;PUT STATUS INTO IOBUFFER
        INC     BX
        DEC     DX
        JNZ     DSTATUS2

;       JMP     UNHOG_DISK
        RET

ERROR_92Y:
        JMP     ERROR_92


;DTURNOFF:  TURN OFF RETRY AND IGNORE ADDRESS CHECK
;

DTURNOFF:
        MOV     BYTE PTR DS:[RETRY_COUNT],0 ;INIT RETRY COUNT
DTURNOFF3:
        INC     BYTE PTR DS:[RETRY_COUNT]   ;INCREMENT RETRY COUNT
        CALL    CMNDINIT                    ;INITIALIZATION COMMAND
        MOV     BYTE PTR DS:[COMMAND1],$EXTEND  ;EXTENDED COMMAND
        MOV     BYTE PTR DS:[COMMAND2],$TURNOFF ;COMMAND = TURN OFF RETRY ...
        CALL    DISKCMDC                    ;SEND COMMAND TO DISK
        CMP     AL,0
        JE      DTURNOFF1                   ;BRANCH IF COMMAND OK

        CMP     AL,60H
        JNZ     DTURNOFF9
        MOV     AX,STACK_BUFFER
        MOV     SP,AX
        CALL    UNHOG_DISK
        CALL    DELAY5
        JMP     PROCESS_NEXT1

DTURNOFF9:
        CMP     BYTE PTR DS:[RETRY_COUNT],16
        JB      DTURNOFF3                   ;BRANCH IF < 16 RETRIES
        JMP     ERROR_94
DTURNOFF1:
        MOV     AL,00H
        CALL    OBSDISK                     ;START OPERATION

;       JMP     UNHOG_DISK
        RET


; DCOPY : COPY SECTOR ON SAME DEVICE
;

DCOPY:
        MOV     BYTE PTR DS:[RETRY_COUNT],0   ; INIT RETRY COUNT
COPYSP1:
        INC     BYTE PTR DS:[RETRY_COUNT]

DCOPY92:
        CALL    CMNDINIT
        MOV     BYTE PTR DS:[COMMAND1],$EXTEND  ;EXTENDED COMMAND
        MOV     BYTE PTR DS:[COMMAND2],$COPY    ;COMMAND = COPY
        CALL    DISKCMND        ;COMMAND/PLATTER/SECTOR ADDRESS ==> DISK
        CMP     AL,0
        JZ      COPYSP2

        CMP     AL,60H          ;ECHO = 60h indicates extended command reject
        JNE     DCOPY9

        MOV     AX,STACK_BUFFER
        MOV     SP,AX
        CALL    UNHOG_DISK
        CALL    DELAY5
        JMP     PROCESS_NEXT1

DCOPY9:
        CMP     BYTE PTR DS:[RETRY_COUNT],16
        JB      RETRYCPY
        JMP     ERROR_94

RETRYCPY:
        TEST    AL,0F0H
        JZ      COPYSP1         ;BRANCH IF ECHO ERROR
        TEST    AL,0FH
        JZ      COPYSP1

        JMP     COPYS11         ;BRANCH IF DPU COPY NOT SUPPORTED

COPYSP2:
        CALL    AKN             ;ACKNOWLEDGEMENT OF VALID SECTOR ADDRESS
        MOV     AL,BYTE PTR DS:[BP].@ADDR21_4
        MOV     BYTE PTR DS:[ADDR1],AL
        MOV     AL,BYTE PTR DS:[BP].@ADDR22_4
        MOV     BYTE PTR DS:[ADDR2],AL
        MOV     AL,BYTE PTR DS:[BP].@ADDR23_4
        MOV     BYTE PTR DS:[ADDR3],AL
        CALL    DISKCMDA        ;'FROM' END ADDRESS ==> DISK
        CMP     AL,0
        JZ      COPYSP2A

        CMP     BYTE PTR DS:[RETRY_COUNT],16  ;BRANCH IF < 16 RETRIES
        JB      COPYSP1
        JMP     ERROR_94

COPYSP2A:
        CALL    AKN

        MOV     AX,WORD PTR DS:[BP].@PLATTER2_4
        MOV     WORD PTR DS:[ACTDEV],AX
        MOV     AL,BYTE PTR DS:[BP].@ADDR31_4
        MOV     BYTE PTR DS:[ADDR1],AL
        MOV     AL,BYTE PTR DS:[BP].@ADDR32_4
        MOV     BYTE PTR DS:[ADDR2],AL
        MOV     AL,BYTE PTR DS:[BP].@ADDR33_4
        MOV     BYTE PTR DS:[ADDR3],AL

        MOV     BYTE PTR DS:[COMMAND1],00H

        MOV     AL,00H
        OUT     0006H,AL            ;SELECT MASTER DISK
        TEST    WORD PTR DS:[ACTDEV],0040H
        JZ      COPYSP2Z
        MOV     AL,01H
        OUT     0006H,AL            ;SELECT SLAVE DISK
COPYSP2Z:
        CALL    DISKCMND        ;'TO' PLATTER/SECTOR ADDRESS ==> DISK
        CMP     AL,0
        JZ      COPYSP2B        ;BRANCH IF NO ECHO ERROR

        CMP     AL,60H          ;ECHO = 60h indicates extended command reject
        JNE     DCOPY91

        MOV     AX,STACK_BUFFER
        MOV     SP,AX
        CALL    UNHOG_DISK
        CALL    DELAY5
        JMP     PROCESS_NEXT1

DCOPY91:
        CMP     BYTE PTR DS:[RETRY_COUNT],16
        JB      COPYSP2C
        JMP     ERROR_94

COPYSP2C:
        INC     BYTE PTR DS:[RETRY_COUNT]
        MOV     AX,WORD PTR DS:[BP].@PLATTER1_4
        MOV     WORD PTR DS:[ACTDEV],AX
        MOV     AL,BYTE PTR DS:[BP].@ADDR11_4
        MOV     BYTE PTR DS:[ADDR1],AL
        MOV     AL,BYTE PTR DS:[BP].@ADDR12_4
        MOV     BYTE PTR DS:[ADDR2],AL
        MOV     AL,BYTE PTR DS:[BP].@ADDR13_4
        MOV     BYTE PTR DS:[ADDR3],AL
        JMP     DCOPY92

COPYSP2B:
        CALL    AKN
        MOV     AL,0
        CALL    OBSDISK         ;START COPYING
        CALL    AKN1            ;COPY ACKNOWLEDGEMENT

;       JMP     UNHOG_DISK
        RET


COPYS1:
        MOV     AL,BYTE PTR DS:[BP].@ADDR23_4
        CMP     AL,BYTE PTR DS:[ADDR3_S1]
        JNE     COPYS1A
        MOV     AL,BYTE PTR DS:[BP].@ADDR22_4
        CMP     AL,BYTE PTR DS:[ADDR2_S1]
        JNE     COPYS1A
        MOV     AL,BYTE PTR DS:[BP].@ADDR21_4
        CMP     AL,BYTE PTR DS:[ADDR1_S1]
        JNE     COPYS1A
        JMP     COPYS2

COPYS1A:
        ADD     BYTE PTR DS:[ADDR3_S1],1
        ADC     BYTE PTR DS:[ADDR2_S1],0
        ADC     BYTE PTR DS:[ADDR1_S1],0

        ADD     BYTE PTR DS:[ADDR3_S2],1
        ADC     BYTE PTR DS:[ADDR2_S2],0
        ADC     BYTE PTR DS:[ADDR1_S2],0
        JMP     COPYSXX

COPYS11:
        MOV     AL,BYTE PTR DS:[BP].@ADDR11_4
        MOV     BYTE PTR DS:[ADDR1_S1],AL
        MOV     AL,BYTE PTR DS:[BP].@ADDR12_4
        MOV     BYTE PTR DS:[ADDR2_S1],AL
        MOV     AL,BYTE PTR DS:[BP].@ADDR13_4
        MOV     BYTE PTR DS:[ADDR3_S1],AL

        MOV     AL,BYTE PTR DS:[BP].@ADDR31_4
        MOV     BYTE PTR DS:[ADDR1_S2],AL
        MOV     AL,BYTE PTR DS:[BP].@ADDR32_4
        MOV     BYTE PTR DS:[ADDR2_S2],AL
        MOV     AL,BYTE PTR DS:[BP].@ADDR33_4
        MOV     BYTE PTR DS:[ADDR3_S2],AL

COPYSXX:
        MOV     AL,BYTE PTR DS:[ADDR1_S1]
        MOV     BYTE PTR DS:[ADDR1],AL
        MOV     AL,BYTE PTR DS:[ADDR2_S1]
        MOV     BYTE PTR DS:[ADDR2],AL
        MOV     AL,BYTE PTR DS:[ADDR3_S1]
        MOV     BYTE PTR DS:[ADDR3],AL

        MOV     AX,WORD PTR DS:[BP].@PLATTER1_4
        MOV     WORD PTR DS:[ACTDEV],AX

        MOV     WORD PTR DS:[SECTOR_COUNT],1

        CALL    GET_BUFFER_PTR
        CALL    DREAD                    ;READ UP TO 4 SECTOR 'FROM' DEVICE

        MOV     AX,WORD PTR DS:[BP].@PLATTER2_4
        MOV     WORD PTR DS:[ACTDEV],AX

        MOV     AL,BYTE PTR DS:[ADDR1_S2]
        MOV     BYTE PTR DS:[ADDR1],AL
        MOV     AL,BYTE PTR DS:[ADDR2_S2]
        MOV     BYTE PTR DS:[ADDR2],AL
        MOV     AL,BYTE PTR DS:[ADDR3_S2]
        MOV     BYTE PTR DS:[ADDR3],AL

        MOV     WORD PTR DS:[SECTOR_COUNT],1
        CALL    GET_BUFFER_PTR

        AND     AH,0FH
        CMP     AH,0DH
        JZ      COPYS3          ;BRANCH IF PHOENIX WRITING TO
        CALL    DWRITE          ;WRITE SECTORS THAT WERE READ TO 'TO' DEVICE
        JMP     COPYS1

COPYS2:
;       JMP     UNHOG_DISK
        RET

; WRITING TO PHOENIX
;

COPYS3:
;       CALL    MSWRTBEG        ;BEGIN MULTI-SECTOR WRITE
        CALL    DWRITE          ;WRITE SECTORS
;       CALL    MSWRTEND        ;END MULTI-SECTOR WRITE
        JMP     COPYS1


;     CMNDINIT: INITIALIZE DISK CONTROLLER TO RECEIVE A
;               COMMAND.  (CKENDV ON 2200T, 4600)
;
;          INITIALIZATION STROBE DATA INDICATES 2200 TYPE TO
;          THE DISK (CURRENTLY, ONLY DDDD FLOPPY CARES):
;               00 -- 2200T
;               01 -- 2200VP
;               02 -- 2200MVP
;               03 -- CS/386
;
;        Disk acknowledgement of initialization indicates the
;        type of drive:
;               Cx -- Diablo/Floppy
;               D0 -- Phoenix, DSDD
;
;        CMNDINIT translates ACTDEV specification as
;        follows:
;
;          For Diablo/Floppy:
;               T/D10 ==> /B10
;               T/D11 ==> /310
;
;          For Phoenix:
;               T/B10 ==> /D10
;               R/B10,R/310 ==> /D10
;               T/310 ==> /D11
;               F/B10,F/310 ==> /D11
;
;          Thus, device type in ACTDEV is 'D' iff Phoenix disk


CMNDINIT:MOV       AL,00H
         OUT       0006H,AL            ;SELECT MASTER DISK
         TEST      WORD PTR DS:[ACTDEV],0040H
         JZ        CMNDINTX
         MOV       AL,01H
         OUT       0006H,AL            ;SELECT SLAVE DISK

CMNDINTX:;CALL      DELAY10             ;DELAY 20 USEC. (MUST NOT PERFORM NEXT
         ;CALL      DELAY10             ;OPERATION TOO SOON AFTER LAST)

;        CALL      DELAY5

         MOV       AL,00H
         OUT       0002H,AL            ;SET CPU BUSY

         CALL      HOG_DISK

CMNDINT3:CALL      WAITRDY             ;WAIT FOR DISK READY
         MOV       AL,01H
         OUT       0004H,AL
         OUT       0008H,AL            ;SET IOB'S TO A0 FOR STRAT OPERATION
         MOV       AL,BYTE PTR DS:[CPU_ID]
         SHL       AL,1
         SHL       AL,1
         SHL       AL,1
         OR        AL,03H              ;CS/386
         OUT       0080H,AL            ;STROBE COMMAND TO DISK
;        CALL      DELAY5
         MOV       CX,50
         CALL      WAITIBS             ;WAIT UP TO 50MS FOR RESPONSE (IBS) FROM DISK
         JC        ERROR_90            ;BRANCH IF NO RESPONSE
         MOV       DX,WORD PTR DS:[ACTDEV]  ;PC'S = DEVICE TYPE & ADDRESS
         CMP       AL,0D0H
         JE        CMNDINT5            ;BRANCH IF PHOENIX TYPE
         AND       AL,0F0H
         CMP       AL,0C0H
         JE        CMNDINT4            ;BRANCH IF FLOPPY/DIABLO TYPE

ERROR_90:MOV       AL,90H              ;DISK HARDWARE ERROR
         JMP       DISKERR


; DIABLO/FLOPPY TYPE DRIVE
;

CMNDINT4:MOV       BL,DH
         AND       BL,0FH
         CMP       BL,03H
         JE        CMNDINT6            ;BRANCH IF DEVICE TYPE = 3 OR B
         CMP       BL,0BH
         JE        CMNDINT6
         AND       DH,0F0H
         OR        DH,0BH              ;CHANGE DEVICE TYPE FROM 'D' TO 'B' IF ADDR IS X0
         TEST      DL,0FH
         JZ        CMNDINT2            ;BRANCH IF ADDRESS IS X0 (REMOVABLE PLATTER
         AND       DH,0F7H             ;DEVICE TYPE = '3' IF ADDRESS IS X1 (FIXED )
         MOV       BL,DL
         AND       BL,0FH
         CMP       BL,01H
         JNE       ERR_48F             ;DIABLO/FLOPPY HAS ONLY 2 PLATTERS (F,R)
         AND       DL,0FEH             ;STRIP PLATTER SPEC. OFF DISK ADDRESS
CMNDINT2:MOV       WORD PTR DS:[ACTDEV],DX   ;STORE TRANSLATED DEVICE TYPE & ADDRESS
CMNDINT6:RET

; PHOENIX TYPE DRIVE
;

CMNDINT5:MOV       BL,DH
         AND       BL,0FH
         CMP       BL,0DH
         JE        CMNDINT6            ;BRANCH IF DEVICE TYPE IS 'D'
         TEST      DL,0FH
         JNZ       ERR_48F             ;ERROR IF LOW 4-BITS OF ADDRESS NOT 0
         MOV       BL,DH
         AND       BL,0F0H
         CMP       BL,20H
         JE        CMNDINT7            ;BRANCH IF 'R' (0=T,1=F,2=R)
         CMP       BL,10H
         JE        CMNDINT8            ;BRANCH IF 'F'
         MOV       BL,DH
         AND       BL,0FH
         CMP       BL,0BH
         JE        CMNDINT7            ;BRANCH IF DEVICE TYPE = 'B'
CMNDINT8:OR        DL,01H              ;PLATTER = FIXED
CMNDINT7:AND       DH,0F0H
         OR        DH,0DH              ;DEVICE TYPE = 'D'
         MOV       WORD PTR DS:[ACTDEV],DX  ;SAVE TRANSLATED DEVICE TYPE & ADDRESS
         RET

ERR_48F: MOV       AL,48H              ;ILLEGAL DEVICE SPECIFICATION
         JMP       DISKERR


;     DISKCMND: SEND COMMAND, PLATTER, AND SECTOR ADDRESS
;               TO DISK (SKADR ON 2200T, 41D3)
;
;         COMMAND FORMAT (DIABLO/FLOPPY READ, WRITE, COMPARE):
;                 BYTE1       BYTE2       BYTE3
;               XX0X XXXX   XXXX XXXX   XXXX XXXX
;               -- - ----------------------------
;               !  !         !
;               !  !                    SECTOR ADDRESS (2600
;               !  !               USES LOW 15 BITS ONLY)
;               !  !
;               !  !                    PLATTER (0 = FIXED, 1 = REMVBL)
;               !
;               !                       COMMAND: 00 = READ
;                                  01 = WRITE
;                                  10 = COMPARE
;
;         COMMAND FORMAT (PHOENIX READ, WRITE, COMPARE):
;                 BYTE1       BYTE2       BYTE3       BYTE4
;               XX0X XXXX   XXXX XXXX   XXXX XXXX   XXXX XXXX
;               -- ------   ---------------------------------
;               !  !         !
;               !  !                    SECTOR ADDRESS (2600
;               !  !               USES LOW 16 BITS ONLY)
;               !  !
;               !  !                    PLATTER (10 = REMOVABLE)
;               !                 (00 = 1ST FIXED)
;               !                 (01 = 2ND FIXED)
;               !                  .
;               !                  .
;               !                 (0E = 15TH FIXED)
;               !
;               !                       COMMAND: 00 = READ
;                                  01 = WRITE
;                                  10 = COMPARE
;
;         COMMAND FORMAT (PHOENIX EXTENDED COMMANDS):
;     BYTE1       BYTE2       BYTE3       BYTE4       BYTE5
;   001X XXXX   XXXX XXXX   XXXX XXXX   XXXX XXXX   XXXX XXXX
;      ------   ---------   ---------------------------------
;      !        !            !
;      !        !                       SECTOR ADDRESS (2600
;      !        !                  USES LOW 16 BITS ONLY)
;      !        !
;      !        !                      EXTENDED COMMAND CODE (E.G., COPY)
;      !
;      !                                PLATTER (10 = REMOVABLE)
;                                 (00 = 1ST FIXED)
;                                 (01 = 2ND FIXED)
;                                  .
;                                  .
;                                 (0E = 15TH FIXED)
;
;          RETURN: AL = 0 IF COMMAND REC'D OK
;                     <>0 IF ECHO ERROR
;
;          NOTE: IF ECHO ERROR OCCURS ON LAST BYTE OF
;                COMMAND, 2600 MUST WAIT FOR ACK TO BE
;                RECEIVED FROM DISK PROCESSOR BEFORE
;                RE-INITIALIZING DISK PROCESSOR.
;                (THIS IS BUG ON 2200)


DISKCMND:
        CALL    DISKCMDC            ;SEND COMMAND TO DISK
        CMP     AL,0
        JE      DISKCMDB            ;SEND SECTOR ADDRESS TO DISK
        JMP     SR@5                ;RETURN IF ECHO ERROR

; SEND COMMAND TO DISK
;

DISKCMDC:
        MOV     AL,00H
        OUT     0004H,AL
        OUT     0008H,AL            ;SET IOB'S 40 FOR OPERATION IN PROGRESS

        MOV     DX,WORD PTR DS:[ACTDEV]
        MOV     AL,DH
        MOV     CL,4
        SHR     AL,CL
        TEST    AL,0FH
        JZ      DISKCMD3            ;BRANCH IF DEVICE TYPE 'T'
        TEST    AL,01H
        JNZ     DISKCMD2

DISKCMD1:
        OR      BYTE PTR DS:[COMMAND1],10H ;PLATTER = 0 (FIXED) OR 1 (REMOVABLE)
DISKCMD2:
        MOV     AL,BYTE PTR DS:[COMMAND1]
        CALL    OBSECHO                    ;SEND COMMAND/PLATTER
        CMP     AL,0
        JNE     DISKCMD5                   ;EXIT IF ECHO ERROR
        TEST    BYTE PTR DS:[COMMAND1],20H
        JZ      DISKCMD5                   ;BRANCH IF NOT EXTENDED COMMAND
        MOV     AL,BYTE PTR DS:[COMMAND2]
        JMP     OBSECHO                    ;SEND 2ND BYTE OF COMMAND

DISKCMD3:
        MOV     DX,WORD PTR DS:[ACTDEV]
        MOV     AL,DH
        AND     AL,0FH
        CMP     AL,0DH
        JE      DISKCMD4                   ;BRANCH IF DEVICE TYPE 'D'
        CMP     AL,03H
        JE      DISKCMD2                   ;BRANCH IF DEVICE TYPE '3'
        JMP     DISKCMD1

DISKCMD4:
        TEST    DL,0FH
        JZ      DISKCMD1                   ;BRANCH IF DEVICE DX0 (REMOVABLE)
        MOV     AL,DL
        AND     AL,0FH
        DEC     AL                         ;CALCULATE PLATTER#
        OR      BYTE PTR DS:[COMMAND1],AL  ;D11 ==> 00, D12 ==> 01, ...
        JMP     DISKCMD2

; SEND SECTOR ADDRESS TO DISK
;

DISKCMDB:
        MOV     DX,WORD PTR DS:[ACTDEV]
        MOV     AL,DH
        AND     AL,0FH
        CMP     AL,0DH
        JE      DISKCMDA            ;BRANCH IF DEVICE TYPE 'D'
        MOV     AL,BYTE PTR DS:[ADDR2]
        AND     AL,7FH              ;STRIP OFF 16TH BIT OF ADDRESS
        JMP     DISKCMD6

DISKCMDA:
        MOV     AL,BYTE PTR DS:[ADDR1]
        CALL    OBSECHO             ;HIGH BYTE OF ADDRESS = 0
        CMP     AL,0
        JNE     DISKCMD5            ;EXIT IF ECHO ERROR
        MOV     AL,BYTE PTR DS:[ADDR2]

DISKCMD6:
        CALL    OBSECHO             ;SEND ADDRESS BYTE TO DISK
        CMP     AL,0
        JNE     DISKCMD5            ;EXIT IF ECHO ERROR
        MOV     AL,BYTE PTR DS:[ADDR3]
        CALL    OBSECHO             ;SEND ADDRESS BYTE TO DISK
        CMP     AL,0
        JE      SR@5                ;RETURN IF NO ECHO ERROR
        MOV     CX,50
        CALL    WAITIBS             ;WAIT FOR ACKNOWLEDGEMENT FROM DISK (MUST
                                      ;   GET BEFORE CAN RE-INITTIALIZE DISK)

        JC      ERR_90F             ;BR, IF NO RESPONE WITHIN 15MS

DISKCMD5:
;       JMP     DELAY5

SR@5:   RET

ERR_90F:MOV     AL,90H
        JMP     DISKERR


;      AKN: GET ACKNOWLEDGEMENT (IBS) FROM DISK OF
;           VALID SECTOR ADDRESS (2200T, 4184)
;
;           ACKNOWLEDGEMENT CODE:
;             AL      = 0 IF OK
;             AL,BIT0 = 1 IF ERROR#98 (ADDRESSING ERROR)
;             AL,BIT1 = 1 IF ERROR#91 (DISK HARDWARE ERROR
;             AL,BIT2 = 1 IF ERROR#94 (FORMAT KEY ENGAGED)
;
;      AKN1: GET ACKNOWLEDGEMENT (IBS) FROM DISK. ON VP,
;            ERROR IF  NOT RECEIVED WITHIN 5SEC.
;            HANGS ON MVP.
;
;           ACKNOWLEDGEMENT CODE:
;            AL     = 0 IF OK
;            AL,BIT0 = 1 IF ERROR#95 (SEEK ERROR)
;            AL,BIT1 = 1 IF ERROR#93 (DISK FORMAT ERROR)
;            AL,BIT2 = 1 IF ERROR#96 (CRC ERROR)

AKN:
        MOV     CX,15
        CALL    WAITIBS                ;WAIT FOR IBS FROM DISK
        JC      ERROR_92
        TEST    AL,1
        JNZ     ERROR_98
        TEST    AL,2
        JNZ     ERROR_91
        JMP     DELAY5                 ;DELAY SINCE MUX. HAS COMMON I/O DATA LINES
        RET

ERROR_98:
        MOV     AL,98H                 ;SECTOR NOT ON DISK (ADDRESSING ERROR)
        JMP     DISKERR
ERROR_91:
        MOV     AL,91H                 ;DISK HARDWARE ERROR
        JMP     DISKERR
ERROR_94:
        MOV     AL,94H                 ;FORMAT KEY ENGAGED
        JMP     DISKERR

AKN1:
        MOV     CX,15
        CALL    WAITIBS                ;WAIT FOR ACK FROM DISK
        JC      ERROR_92
AKN1E:
        TEST    AL,1
        JNZ     ERROR_95
        TEST    AL,2
        JNZ     ERROR_93
        TEST    AL,4
        JNZ     ERROR_96
        RET

ERROR_93:
        MOV     AL,93H                 ;DISK FORMAT ERROR
        JMP     DISKERR
ERROR_96:
        MOV     AL,96H                 ;CRC ERROR
        JMP     DISKERR
ERROR_95:
        MOV     AL,95H                 ;SEEK ERROR
        JMP     DISKERR
ERR_92:
ERROR_92:
        MOV     AL,92H                 ;DISK HARDWARE ERROR
        JMP     DISKERR


;      OBSECHO: STROBES (OBS) THE CONTENTS OF F0
;              AND CHECKS ECHO (IBS)  (STRC ON 2200T,41E8)
;
;           ENTRY: AL = CHARACTER TO STROBE
;
;           RETURN: AL = 0 IF ECHO = CHARACTER
;                     <> 0 IF ECHO INCORRECT CHARACTER

OBSECHO:
        PUSH    AX
        CALL    OBSDISK
        MOV     CX,15
        CALL    WAITIBS
        JC      ERROR_92
        MOV     BL,AL
        POP     AX
        XOR     AL,BL
        RET

; OBSDISK: SAME AS 'OBSTROBD' EXCEPT IOB'S RESTORED TO HEX(40)
;          IF BREAKPOINT OCCURS.   OUTPUTS K WITH AN OBS.
;

OBSDISK:
        CALL    WAITRDY
        OUT     0080H,AL
;       CALL    DELAY5
        RET

; ADD_SECTOR_ADDR: INCREMENT SECTOR ADDRESS
;

ADD_SECTOR_ADDR:
        ADD     BYTE PTR DS:[ADDR3],1          ;INCREMENT SECTOR ADDRESS
        ADC     BYTE PTR DS:[ADDR2],0
        ADC     BYTE PTR DS:[ADDR1],0
        RET

; WAITRDY: WAIT DISK READY
;

WAITRDY:PUSH    AX

WAITRDY1:
        CALL    BREAK               ;BREAK WHEN NOT READY
        IN      AL,0084H            ;READ STATUS
        TEST    AL,20H              ;CHECK READY BIT
        JNZ     WAITRDY1            ;BR, IF NOT READY

        POP     AX
        RET

; BREAK: BREAK FOR CHECK INTERUPT FROM CS/386
;

BREAK:
        CMP     BYTE PTR DS:[RESET_PRT_FLAG],0FFH
        JE      BREAK6
        CALL    CLEAR_BUFFER
        MOV     BYTE PTR DS:[RESET_PRT_FLAG],0FFH
BREAK6:
        CMP     BYTE PTR DS:[RESET_FLAG],0FFH
        JE      BREAK1
        CMP     BYTE PTR DS:[RESET_FLAG],80H
        JE      BREAK4
        PUSH    AX
        PUSH    BX
        XOR     BX,BX
        MOV     BL,BYTE PTR DS:[RESET_FLAG]
        MOV     BYTE PTR DS:[RESET_FLAG],0FFH
        CMP     BL,BYTE PTR DS:[ACT_PART_NO]
        JNE     BREAK2
        CALL    RESET_DSK
;       CALL    UNHOG_DISK
        MOV     BX,STACK_BUFFER
        MOV     SP,BX
        JMP     PROCESS_NEXT
BREAK2: CMP     BYTE PTR DS:[ACT_PART_NO],0FFH
        JNE     BREAK3
        CALL    RESET_DSK
        POP     BX
        POP     AX
        RET
BREAK3:
        SHL     BX,1
        SHL     BX,1
        SHL     BX,1
        SHL     BX,1
        SHL     BX,1
        ADD     BX,RCB
        MOV     BYTE PTR DS:[BX].@STATUS,0
        POP     BX
        POP     AX
BREAK1:
        RET
BREAK4:
        MOV    WORD PTR DS:[PRT_START],PRT_BUFFER_START
        MOV    WORD PTR DS:[PRT_END],PRT_BUFFER_END
        MOV    CX,40H
        MOV    BX,RCB
BREAK5:
        MOV    BYTE PTR DS:[BX].@STATUS,0
        ADD    BX,20H
        LOOP   BREAK5
        CALL   RESET_DSK
        MOV    BYTE PTR DS:[RESET_FLAG],0FFH
        MOV    BX,STACK_BUFFER
        MOV    SP,BX
        JMP    PROCESS_NEXT


; WAITIBS: WAIT DATA FROM DISK
;
;     ENTRY:   CX = DELAY COUNT (MS)
;     RETURN:  AL = DATA FROM DISK
;              CARRY SET IF NO RESPONE

WAITIBS:
        PUSH    CX
        CALL    WAITRDY
        POP     CX
        MOV     DX,0300H

        MOV     AL,01H
        OUT     0002H,AL            ;SET CPU READY FOR ALLOW INPUT

WAIT_AGAIN:
        IN      AL,0084H            ;5
        TEST    AL,02H              ;3
        JNZ     WAIT_OK             ;3
        DEC     DX                  ;2
        JNZ     WAIT_AGAIN          ;7
        MOV     DX,0300H            ;
        DEC     CX
        JNZ     WAIT_AGAIN          ;

        JMP     WAIT_ERROR          ;DECREMENT DELAY COUNT

WAIT_OK:
        NOP
        NOP
        NOP
        NOP
        NOP
        NOP
        NOP
        NOP
        MOV     AL,00H
        OUT     0002H,AL            ;SET CPU BUSY
        IN      AL,0082H            ;GET DATA

        CLC                         ;IBS RECIEVE
        RET

WAIT_ERROR:
        MOV     AL,00H
        OUT     0002H,AL            ;SET CPU BUSY

        STC
        RET


; AL CONTAINS ERROR CODE
;

DISKERR:
        MOV     BYTE PTR DS:[BP].@ERROR,AL
        CALL    UNHOG_DISK
        MOV     AX,STACK_BUFFER
        MOV     SP,AX
        JMP     PROCESS_NEXT


; DELAY10:   DELAY 10  US
; DELAY5:    DELAY 5   US

DELAY10:CALL    DELAY5            ; 60 + 11 (call)
        CALL    DELAY5            ; 60
        RET                       ; 11    (142 FOR 10 US)

DELAY5: PUSH    AX                ; 3 + 11 (call)
        POP     AX                ; 5
        PUSH    AX                ; 3
        POP     AX                ; 5
        PUSH    AX                ; 3
        POP     AX                ; 5
        PUSH    AX                ; 3
        POP     AX                ; 5
        NOP                       ; 2
        NOP                       ; 2
        NOP                       ; 2
        RET                       ; 11    (60 FOR 5 US)

DELAY1MS:
        PUSH    CX                ; 3 + 11 (call)
        MOV     CX,5EH            ; 2
DELAYX: CALL    DELAY10           ; 118
        DEC     CX                ; 2
        JNZ     DELAYX            ; 8
        POP     CX                ; 5
        RET                       ; 11    (11992 FOR 1 MS)

; CS/386 TO 286 INTERRUPR SEVICE ROUTINE
;

H286_I: STI
        IRET

; PRINTER INTERRUPT SEVICE ROUTINE
;

PRINTER_I:
        STI
        IRET

; 8255 DATA OUT INTERRUPT SEVICE ROUTINE
;

DATA_OUT_I:
        STI
        IRET

; 8255 DATA IN INTERRUPT SEVICE ROUTINE
;     RETURN:  AL = DATA FROM DISK
;

DATA_IN_I:
        STI
        IRET

;HOG_DISK : SEND REQUEST TO DISK FOR HOG DISK
;

HOG_DISK:
        PUSH    AX
        MOV     AL,01H
        OUT     0000H,AL
        POP     AX
        RET

;UNHOG_DISK : SEND REQUEST TO DISK FOR UNHOG DISK
;

UNHOG_DISK:
        PUSH    AX
        MOV     AL,00H
        OUT     0000H,AL
        POP     AX
        RET

;GET_BUFFER_PTR : GET BUFFER POINTER ACCORDING TO PART_NO
;

GET_BUFFER_PTR:
        PUSH    AX
        PUSH    BX
        MOV     BX,BASE                         ;POINT TO BUFFER BASE OF PART_NO
        XOR     AX,AX
        MOV     AL,BYTE PTR DS:[PART_NO]
        SHL     AX,1
        SHL     AX,1
        ADD     BX,AX
        MOV     AX,WORD PTR DS:[BX]             ;GET SEGMENT OF BUFFER PTR
        MOV     ES,AX
        MOV     AX,WORD PTR DS:[BX+2]           ;GET OFFSET OF BUFFER PTR
        MOV     WORD PTR DS:[BUFFER_PTR],AX
        POP     BX
        POP     AX
        RET

;SET_NEXT_BUFFER_PTR : ADD 256 OF CURRENT BUFFER POINTER
;

SET_NEXT_BUFFER_PTR:
        PUSH    AX
        ADD     WORD PTR DS:[BUFFER_PTR],256
        CMP     WORD PTR DS:[BUFFER_PTR],0
        JNE     SET_RET
        MOV     AX,ES
        ADD     AX,1000H
        CMP     AX,4000H
        JNE     SET_1
        MOV     AX,0
        MOV     WORD PTR DS:[BUFFER_PTR],8000H
SET_1:
        MOV     ES,AX
SET_RET:
        POP     AX
        RET

;ADD_BUFFER_PTR : ADD 1 OF CURRENT BUFFER POINTER
;

ADD_BUFFER_PTR:
        PUSH    AX
        INC     WORD PTR DS:[BUFFER_PTR]
        CMP     WORD PTR DS:[BUFFER_PTR],0
        JNE     ADD_RET
        MOV     AX,ES
        ADD     AX,1000H
        CMP     AX,4000H
        JNE     ADD_1
        MOV     AX,0
        MOV     WORD PTR DS:[BUFFER_PTR],8000H
ADD_1:
        MOV     ES,AX
ADD_RET:
        POP     AX
        RET

; RESET_DSK: RESET DISK
;

RESET_DSK:
        PUSH    AX
        MOV     AL,0
        OUT     000AH,AL
        NOP
        NOP
        NOP
        NOP
        NOP
        MOV     AL,1
        OUT     000AH,AL
        NOP
        NOP
        NOP
        NOP
        NOP
        MOV     AL,0
        OUT     000AH,AL
        POP     AX
        RET

; FLASH_LED : FLASH LED
;

FLASH_LED:
        PUSH    AX
        PUSH    CX
        MOV     AL,01H
        OUT     000EH,AL
        MOV     CX,300H
        LOOP    $
        MOV     AL,00H
        OUT     000EH,AL
        MOV     CX,300H
        LOOP    $
        POP     CX
        POP     AX
        RET
